/*****************************************************************************
 *            ASM: a minimal assembler for educational purpose               *
 ***************************************************************************** 
 *                                                                           *
 * Copyright (c) 2010 Christophe Alias                                       *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License as published by the     *
 * Free Software Foundation; either version 2 of the License, or (at your    * 
 * option) any later version.                                                *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General  *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License along   *
 * with this program; if not, write to the Free Software Foundation, Inc.,   *
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                 *
 *                                                                           *
 * Written by Christophe Alias, Christophe.Alias@ens-lyon.fr                 *
 *                                                                           *
 *****************************************************************************/
#include "code_generation.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Target binary code */

unsigned short int code[MAX_INSTRUCTIONS];
int cp; //current instruction

/* Label declaration */

char* label_id[MAX_INSTRUCTIONS]; //id 
unsigned short int label_address[MAX_INSTRUCTIONS]; //address
int current_label;

/* Label use */

char* label_ldi_lo[MAX_INSTRUCTIONS];
char* label_ldi_hi[MAX_INSTRUCTIONS];
char* label_beq[MAX_INSTRUCTIONS];
char* label_j[MAX_INSTRUCTIONS];

#define ADD 0 << 13
#define SUB 1 << 13
#define LD  2 << 13
#define ST  3 << 13
#define BEQ 4 << 13
#define LDI 5 << 13
#define JA  6 << 13
#define J   7 << 13

#define set_r1(n) (n) << 10
#define set_r2(n) (n) << 7
#define set_r3(n) (n) << 4

void init()
{
  int i;
  cp = -1;
  for(i=0; i<MAX_INSTRUCTIONS; i++)
    {
      code[i] = 0;
      label_ldi_lo[i] = strdup("(none)");
      label_ldi_hi[i] = strdup("(none)");
      label_beq[i] = strdup("(none)");
      label_j[i] = strdup("(none)");
    }

  current_label = -1;
}

void next(int* cp)
{
  (*cp)++;
  if((*cp)>=MAX_INSTRUCTIONS)
    {
      printf("Input program is too big\n");
      exit(1);
    }
}

void emit_add(unsigned short int r_target, 
	      unsigned short int r1, 
	      unsigned short int r2)
{
  next(&cp);
  code[cp] = ADD | set_r1(r_target) | set_r2(r1) | set_r3(r2);
}

void emit_sub(unsigned short int r_target, 
	      unsigned short int r1, 
	      unsigned short int r2)
{
  next(&cp);
  code[cp] = SUB | set_r1(r_target) | set_r2(r1) | set_r3(r2);
}

void emit_ldi(unsigned short int r_target, 
	      unsigned short int value)
{
  next(&cp);
  code[cp] = LDI | set_r1(r_target) | value;
}


void emit_ldi_label_lo(unsigned short int r_target, char* label)
{
  next(&cp);
  code[cp] = LDI | set_r1(r_target);
  label_ldi_lo[cp] = strdup(label);
}

void emit_ldi_label_hi(unsigned short int r_target, char* label)
{
  next(&cp);
  code[cp] = LDI | set_r1(r_target);
  label_ldi_hi[cp] = strdup(label);
}

void emit_ld(unsigned short int r_target, 
	     unsigned short int r_base, 
	     unsigned short int offset)
{
  next(&cp);
  code[cp] = LD | set_r1(r_target) | set_r2(r_base) | offset;
}


void emit_ld_neg(unsigned short int r_target, 
		 unsigned short int r_base, 
		 unsigned short int offset)
{
  next(&cp);
  short int neg = -1*offset;
  code[cp] = LD | set_r1(r_target) | set_r2(r_base) | (neg & 127);
}


void emit_st(unsigned short int r_source, 
	     unsigned short int r_base, 
	     unsigned short int offset)
{
  next(&cp);
  code[cp] = ST | set_r1(r_source) | set_r2(r_base) | offset;
}

void emit_st_neg(unsigned short int r_source, 
		 unsigned short int r_base, 
		 unsigned short int offset)
{
  next(&cp);
  short int neg = -1*offset;
  code[cp] = ST | set_r1(r_source) | set_r2(r_base) | (neg & 127);
}

void emit_beq(unsigned short int r1, 
	      unsigned short int r2, char* label)
{
  next(&cp);
  code[cp] = BEQ | set_r1(r1) | set_r2(r2);
  label_beq[cp] = strdup(label);
}

void emit_ja(unsigned short int r1, 
	     unsigned short int r2)
{
  next(&cp);
  code[cp] = JA | set_r1(r1) | set_r2(r2);
}

void emit_j(char* label)
{
  next(&cp);
  code[cp] = J;
  label_j[cp] = strdup(label);
}

void emit_data_num1(unsigned short int value)
{
  next(&cp);
  code[cp] = value;
}

void emit_data_num2(unsigned short int value0, unsigned short int value1)
{
  next(&cp);
  code[cp] = (value1 << 8) | (255 & value0);
}

void emit_data_string(char* s)
{
  int i;
  for(i=0; i<strlen(s); i+=2)
    {
      next(&cp);
      code[cp] = ((char) *(s+i)) << 8;
      if (i+1<strlen(s))
	code[cp] = code[cp] | (255 & ((char) *(s+i+1)));
    }
}

void define_label(char* label)
{
  current_label++;
  label_id[current_label] = strdup(label);
  label_address[current_label] = cp+1;
}

void fix_labels()
{
  int i,j,label_found;
  for(i=0; i<MAX_INSTRUCTIONS; i++)
    //For each instruction using a label
    {
      /*
       * LDI LO(label)
       */
      if(strcmp(label_ldi_lo[i],"(none)") != 0)
	{
	  //Get the label
	  label_found = 0;
	  for(j=0; j<=current_label; j++)
	    if(strcmp(label_id[j],label_ldi_lo[i]) == 0)
	      {
		//Compute the offset
		unsigned short int offset = label_address[j];	    
		//Fix the instruction
		code[i] = code[i] | (255 & offset); //Lo part
		label_found = 1;
	      }
	  if(label_found == 0)
	    {
	      printf("Error: undefined label (ldi lo): %s\n",label_ldi_lo[i]);
	      exit(1);
	    }
	}

      /*
       * LDI HI(label)
       */
      if(strcmp(label_ldi_hi[i],"(none)") != 0)
	{
	  //Get the label
	  label_found = 0;
	  for(j=0; j<=current_label; j++)
	    if(strcmp(label_id[j],label_ldi_hi[i]) == 0)
	      {
		//Compute the offset
		unsigned short int offset = label_address[j];	    
		//Fix the instruction
		code[i] = code[i] | (255 & (offset>>8)); //Hi part
		label_found = 1;
	      }
	  if(label_found == 0)
	    {
	      printf("Error: undefined label (ldi hi): %s\n",label_ldi_hi[i]);
	      exit(1);
	    }
	}

      /*
       * BEQ
       */
      if(strcmp(label_beq[i],"(none)") != 0)
	{
	  //Get the label
	  label_found = 0;
	  for(j=0; j<=current_label; j++)
	    if(strcmp(label_id[j],label_beq[i]) == 0)
	      {
		//Compute the offset
		short int offset = label_address[j] - i - 1;	    

		if(!((offset >= -128) && (offset <= 127)))
		  {
		    printf("Error: beq is too far (delta=%d)\n",offset);
		    exit(1);
		  }

		//Fix the instruction
		code[i] = code[i] | (127 & offset);
		label_found = 1;
	      }
	  if(label_found == 0)
	    {
	      printf("Error: undefined label (beq): %s\n",label_beq[i]);
	      exit(1);
	    }
	}

      /*
       * J
       */
      if(strcmp(label_j[i],"(none)") != 0)
	//For each instruction using a label (J)
	{
	  //Get the label
	  label_found = 0;
	  for(j=0; j<=current_label; j++)
	    if(strcmp(label_id[j],label_j[i]) == 0)
	      {
		//Compute the offset
		unsigned short int offset = label_address[j];
		//Fix the instruction
		code[i] = code[i] | ((0x1FFF) & offset);
		label_found = 1;
	      }
	  if(label_found == 0)
	    {
	      printf("Error: undefined label (j): %s\n",label_j[i]);
	      exit(1);
	    }
	}
    }

}	       

int min(int x, int y)
{
  if(x<y) return x;
  else return y;
}

void emit_code(char* file_name)
{
  int i,j;
  char file_name_lo[80];
  char file_name_hi[80];
  //Get output file names (lo and hi)
  for(i=0; i<strlen(file_name); i++)
    {
      if(file_name[i] == '.')
	{
	  file_name_lo[i] = '\0';
	  file_name_hi[i] = '\0';
	  strcat(file_name_lo,"_lo.ram");
	  strcat(file_name_hi,"_hi.ram");
	  break;
	}
      file_name_lo[i] = file_name[i];
      file_name_hi[i] = file_name[i];
    }

  //Emit "lo" part
  FILE* output_file = fopen(file_name_lo,"w");
  for(i=0; i<=cp; i+=16)
    {
      fprintf(output_file,"%04X:",i);
      for(j=i; j<min(i+16,cp+1); j++)
	fprintf(output_file,"%02X",code[j] & 255);
      fprintf(output_file,"\n");
    }
  fclose(output_file);

  //Emit "hi" part
  output_file = fopen(file_name_hi,"w");
  for(i=0; i<=cp; i+=16)
    {
      fprintf(output_file,"%04X:",i);
      for(j=i; j<min(i+16,cp+1); j++)
	fprintf(output_file,"%02X",(code[j] >> 8) & 255);
      fprintf(output_file,"\n");
    }
  fclose(output_file);

  for(i=0; i<=cp; i++)
    printf("%04X\n",code[i]);

}
